@@ -37,63 +37,64 @@
    1)	July 30 1994 : For unimodal data, gave "infinite loop"  (end of code)
    2)	Oct  31 2003 : Yong Lu <lyongu+@cs.cmu.edu> : ")" typo in Fortran
                        gave wrong result (larger dip than possible) in some cases
-   $Id: dip.c,v 1.21 2003/11/13 07:58:54 maechler Exp $
+   $Id: dip.c,v 1.24 2011/05/25 12:38:39 maechler Exp $
 */
 
 #include <R.h>
 
 /* Subroutine */
-void diptst(double *x, Sint *n,
+void diptst(double *x, Sint *n_,
 	    double *dip, Sint *lo_hi, Sint *ifault,
-	    Sint *gcm, Sint *lcm, Sint *mn, Sint *mj, Sint *debug)
+	    Sint *gcm, Sint *lcm, Sint *mn, Sint *mj,
+	    Sint *min_is_0, Sint *debug)
 {
-#define low  lo_hi[0]
-#define high lo_hi[1]
+#define low   lo_hi[0]
+#define high  lo_hi[1]
+#define l_gcm lo_hi[2]
+#define l_lcm lo_hi[3]
 
     /* Local variables */
-    int gcmi1, gcmix,  lcmiv, lcmiv1, mnj, mnmnj, mjk, mjmjk,
-	ic, icv, icva, icx, icxa, ig, ih, iv, ix,  j, jb, je, jr,  k, kb, ke, kr;
-    double dip_l, dip_u, dipnew, d, dx, t, temp, C;
-    int N = *n;
+    int mnj, mnmnj, mjk, mjmjk, ig, ih, iv, ix,  i, j, k;
+    double dip_l, dip_u, dipnew;
+    int n = *n_;
 
-    /* Parameter adjustments, so I can do "as with index 1" : x[1]..x[N] */
+    /* Parameter adjustments, so I can do "as with index 1" : x[1]..x[n] */
     --mj;    --mn;
     --lcm;   --gcm;
     --x;
 
 /*-------- Function Body ------------------------------ */
 
-    *ifault = 1;    if (N <= 0) return;
+    *ifault = 1;    if (n <= 0) return;
     *ifault = 0;
 
 /* Check that X is sorted --- if not, return with  ifault = 2*/
 
-    *ifault = 2;    for (k = 2; k <= N; ++k) if (x[k] < x[k - 1]) return;
+    *ifault = 2;    for (k = 2; k <= n; ++k) if (x[k] < x[k - 1]) return;
     *ifault = 0;
 
 /* Check for all values of X identical, */
-/*     and for 1 <= N < 4. */
+/*     and for 1 <= n < 4. */
 
 /* LOW contains the index of the current estimate  of the lower end
    of the modal interval, HIGH contains the index for the upper end.
 */
-    low = 1;	high = N; /*-- IDEA:  *xl = x[low];    *xu = x[high]; --*/
+    low = 1;	high = n; /*-- IDEA:  *xl = x[low];    *xu = x[high]; --*/
 
-    if (N < 4 || x[N] == x[1]) {
-	*dip = 0.;	 return;
-    }
+/* M.Maechler -- speedup: it saves many divisions by n when we just work with
+ * (n * dip) everywhere but the very end! */
+    *dip = (*min_is_0) ? 0. : 1.;
+    if (n < 2 || x[n] == x[1])		goto L_END;
 
-/* M.Maechler -- speedup: it saves many divisions by N when we just work with
- * (N * dip) everywhere but the very end! */
-    *dip = 1.;
+    if(*debug)
+	Rprintf("dip() in C: n = %d; starting with  2N*dip = %g.\n",
+		n, *dip);
 
-    if(*debug) Rprintf("dip() in C: N = %d; starting with N*dip = 1.\n", N);
-
-/* Establish the indices   mn[1..N]  over which combination is necessary
+/* Establish the indices   mn[1..n]  over which combination is necessary
    for the convex MINORANT (GCM) fit.
 */
     mn[1] = 1;
-    for (j = 2; j <= N; ++j) {
+    for (j = 2; j <= n; ++j) {
 	mn[j] = j - 1;
 	while(1) {
 	  mnj = mn[j];
@@ -105,16 +106,16 @@ void diptst(double *x, Sint *n,
 	}
     }
 
-/* Establish the indices   mj[1..N]  over which combination is necessary
+/* Establish the indices   mj[1..n]  over which combination is necessary
    for the concave MAJORANT (LCM) fit.
 */
-    mj[N] = N;
-    for (k = N - 1; k >= 1; k--) {
+    mj[n] = n;
+    for (k = n - 1; k >= 1; k--) {
 	mj[k] = k + 1;
 	while(1) {
 	  mjk = mj[k];
 	  mjmjk = mj[mjk];
-	  if (mjk == N ||
+	  if (mjk == n ||
 	      ( x[k]  - x[mjk]) * (mjk - mjmjk) <
 	      (x[mjk] - x[mjmjk]) * (k - mjk)) break;
 	  mj[k] = mjmjk;
@@ -124,38 +125,48 @@ void diptst(double *x, Sint *n,
 /* ----------------------- Start the cycling. ------------------------------- */
 LOOP_Start:
 
-    if(*debug) Rprintf( "'dip':LOOP-BEGIN: low, high = %5ld,%5ld\n", low,high);
-
-/* Collect the change points for the GCM from HIGH to LOW. */
-
+    /* Collect the change points for the GCM from HIGH to LOW. */
     gcm[1] = high;
-    for(ic = 1; gcm[ic] > low; ic++)
-	gcm[ic+1] = mn[gcm[ic]];
-    icx = ic;
-
-/* Collect the change points for the LCM from LOW to HIGH. */
+    for(i = 1; gcm[i] > low; i++)
+	gcm[i+1] = mn[gcm[i]];
+    ig = l_gcm = i; // l_gcm == relevant_length(GCM)
+    ix = ig - 1; //  ix, ig  are counters for the convex minorant.
 
+    /* Collect the change points for the LCM from LOW to HIGH. */
     lcm[1] = low;
-    for(ic = 1; lcm[ic] < high; ic++)
-	lcm[ic+1] = mj[lcm[ic]];
-    icv = ic;
+    for(i = 1; lcm[i] < high; i++)
+	lcm[i+1] = mj[lcm[i]];
+    ih = l_lcm = i; // l_lcm == relevant_length(LCM)
+    iv = 2; //  iv, ih  are counters for the concave majorant.
 
-    ig = icx; /* ICX, IX, IG are counters for the convex minorant, */
-    ih = icv; /* ICV, IV, IH are counters for the concave majorant. */
-    ix = icx - 1;    iv = 2;
+    if(*debug) {
+	Rprintf("'dip': LOOP-BEGIN: 2n*D= %-8.5g  [low,high] = [%3d,%3d]", *dip, low,high);
+	if(*debug >= 3) {
+	    Rprintf(" :\n gcm[1:%d] = ", l_gcm);
+	    for(i = 1; i <= l_gcm; i++) Rprintf("%d%s", gcm[i], (i < l_gcm)? ", " : "\n");
+	    Rprintf(" lcm[1:%d] = ", l_lcm);
+	    for(i = 1; i <= l_lcm; i++) Rprintf("%d%s", lcm[i], (i < l_lcm)? ", " : "\n");
+	} else { // debug <= 2 :
+	    Rprintf("; l_lcm/gcm = (%2d,%2d)\n", l_lcm,l_gcm);
+	}
+    }
 
 /*	Find the largest distance greater than 'DIP' between the GCM and
  *	the LCM from LOW to HIGH. */
 
-    d = 0.;
-    if (icx != 2 || icv != 2) {
+    // FIXME: <Rconfig.h>  should provide LDOUBLE or something like it
+    long double d = 0.;// <<-- see if this makes 32-bit/64-bit difference go..
+    if (l_gcm != 2 || l_lcm != 2) {
+	if(*debug) Rprintf("  while(gcm[ix] != lcm[iv]) :%s",
+			   (*debug >= 2) ? "\n" : " ");
       do { /* gcm[ix] != lcm[iv]  (after first loop) */
-	  gcmix = gcm[ix];
-	  lcmiv = lcm[iv];
+	  long double dx;
+	  int gcmix = gcm[ix],
+	      lcmiv = lcm[iv];
 	  if (gcmix > lcmiv) {
 	      /* If the next point of either the GCM or LCM is from the LCM,
 	       * calculate the distance here. */
-	      gcmi1 = gcm[ix + 1];
+	      int gcmi1 = gcm[ix + 1];
 	      dx = (lcmiv - gcmi1 + 1) -
 		  (x[lcmiv] - x[gcmi1]) * (gcmix - gcmi1)/(x[gcmix] - x[gcmi1]);
 	      ++iv;
@@ -163,90 +174,105 @@ LOOP_Start:
 		  d = dx;
 		  ig = ix + 1;
 		  ih = iv - 1;
+		  if(*debug >= 2) Rprintf(" L(%d,%d)", ig,ih);
 	      }
 	  }
 	  else {
 	      /* If the next point of either the GCM or LCM is from the GCM,
 	       * calculate the distance here. */
-	      lcmiv1 = lcm[iv - 1];
+	      int lcmiv1 = lcm[iv - 1];
 /* Fix by Yong Lu {symmetric to above!}; original Fortran: only ")" misplaced! :*/
 	      dx = (x[gcmix] - x[lcmiv1]) * (lcmiv - lcmiv1) /
-		  (x[lcmiv] - x[lcmiv1])- (gcmix - lcmiv1 - 1);
+		   (x[lcmiv] - x[lcmiv1])- (gcmix - lcmiv1 - 1);
 	      --ix;
 	      if (dx >= d) {
 		  d = dx;
 		  ig = ix + 1;
 		  ih = iv;
+		  if(*debug >= 2) Rprintf(" G(%d,%d)", ig,ih);
 	      }
 	  }
-
 	  if (ix < 1)	ix = 1;
-	  if (iv > icv)	iv = icv;
+	  if (iv > l_lcm)	iv = l_lcm;
+	  if(*debug) {
+	      if(*debug >= 2) Rprintf(" --> ix = %d, iv = %d\n", ix,iv);
+	      else Rprintf(".");
+	  }
       } while (gcm[ix] != lcm[iv]);
+      if(*debug && *debug < 2) Rprintf("\n");
     }
-    else { /* icx or icv == 2 */
-      d = 1.;
+    else { /* l_gcm or l_lcm == 2 */
+	d = (*min_is_0) ? 0. : 1.;
+	if(*debug)
+	    Rprintf("  ** (l_lcm,l_gcm) = (%d,%d) ==> d := %g\n", l_lcm, l_gcm, d);
     }
 
     if (d < *dip)	goto L_END;
 
 /*     Calculate the DIPs for the current LOW and HIGH. */
+    if(*debug) Rprintf("  calculating dip ..");
+
+    int j_best, j_l = -1, j_u = -1;
 
     /* The DIP for the convex minorant. */
-
     dip_l = 0.;
-    if (ig != icx) {
-      icxa = icx - 1;
-      for (j = ig; j <= icxa; ++j) {
-	temp = 1.;
-	jb = gcm[j + 1];
-	je = gcm[j];
+    for (j = ig; j < l_gcm; ++j) {
+	double max_t = 1.;
+	int j_ = -1, jb = gcm[j + 1], je = gcm[j];
 	if (je - jb > 1 && x[je] != x[jb]) {
-	  C = (je - jb) / (x[je] - x[jb]);
-	  for (jr = jb; jr <= je; ++jr) {
-	    t = (jr - jb + 1) - (x[jr] - x[jb]) * C;
-	    if (temp < t)
-		temp = t;
+	  double C = (je - jb) / (x[je] - x[jb]);
+	  for (int jj = jb; jj <= je; ++jj) {
+	    double t = (jj - jb + 1) - (x[jj] - x[jb]) * C;
+	    if (max_t < t) {
+		max_t = t; j_ = jj;
+	    }
 	  }
 	}
-	if (dip_l < temp)
-	    dip_l = temp;
-      }
+	if (dip_l < max_t) {
+	    dip_l = max_t; j_l = j_;
+	}
     }
 
     /* The DIP for the concave majorant. */
-
     dip_u = 0.;
-    if (ih != icv) {
-      icva = icv - 1;
-      for (k = ih; k <= icva; ++k) {
-	temp = 1.;
-	kb = lcm[k];
-	ke = lcm[k + 1];
-	if (ke - kb > 1 && x[ke] != x[kb]) {
-	  C = (ke - kb) / (x[ke] - x[kb]);
-	  for (kr = kb; kr <= ke; ++kr) {
-	    t = (x[kr] - x[kb]) * C - (kr - kb - 1);
-	    if (temp < t)
-		temp = t;
+    for (j = ih; j < l_lcm; ++j) {
+	double max_t = 1.;
+	int j_ = -1, jb = lcm[j], je = lcm[j + 1];
+	if (je - jb > 1 && x[je] != x[jb]) {
+	  double C = (je - jb) / (x[je] - x[jb]);
+	  for (int jj = jb; jj <= je; ++jj) {
+	    double t = (x[jj] - x[jb]) * C - (jj - jb - 1);
+	    if (max_t < t) {
+		max_t = t; j_ = jj;
+	    }
 	  }
 	}
-	if (dip_u < temp)
-	    dip_u = temp;
-      }
+	if (dip_u < max_t) {
+	    dip_u = max_t; j_u = j_;
+	}
     }
 
+    if(*debug) Rprintf(" (dip_l, dip_u) = (%g, %g)", dip_l, dip_u);
+
     /* Determine the current maximum. */
-    dipnew = (dip_u > dip_l) ? dip_u : dip_l;
-    if (*dip < dipnew)
+    if(dip_u > dip_l) {
+	dipnew = dip_u; j_best = j_u;
+    } else {
+	dipnew = dip_l; j_best = j_l;
+    }
+    if (*dip < dipnew) {
 	*dip = dipnew;
+	if(*debug)
+	    Rprintf(" -> new larger dip %g (j_best = %d)\n", dipnew, j_best);
+    }
+    else if(*debug) Rprintf("\n");
 
     /*--- The following if-clause is NECESSARY  (may loop infinitely otherwise)!
       --- Martin Maechler, Statistics, ETH Zurich, July 30 1994 ---------- */
     if (low == gcm[ig] && high == lcm[ih]) {
       if(*debug)
 	Rprintf("No improvement in  low = %ld  nor  high = %ld --> END\n",
-	       low, high);
+		low, high);
     } else {
 	low  = gcm[ig];
 	high = lcm[ih];	   goto LOOP_Start; /* Recycle */
@@ -257,7 +283,7 @@ L_END:
     /* do this in the caller :
      *   *xl = x[low];  *xu = x[high];
      * rather return the (low, high) indices -- automagically via lo_hi[]  */
-    *dip /= (2*N);
+    *dip /= (2*n);
     return;
 } /* diptst */
 #undef low
